use crate::vm::dm_type::DmPrimitiveType;
use crate::vm::mem::DmAllocObject;
use std::fmt::Debug;
use std::rc::Rc;

#[derive(Debug, PartialEq, Eq)]
pub enum DmObjectType {
    Interface(DmInterface),
    Implementation(DmImplementation),
}

#[derive(Debug, Eq)]
pub struct DmFn {
    pub fqn: String,
    pub short_name: String,
    pub implements: Option<DmVirtualMethod>,
    pub byte_code: Vec<u8>,
    pub num_registers: usize,
}

impl PartialEq for DmFn {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmInterface {
    pub fqn: String,
    short_name: String,
    properties: Vec<Rc<DmProperty>>,
    virtual_methods: Vec<Rc<DmVirtualMethod>>,
    impl_methods: Vec<Rc<DmFn>>,
}

impl PartialEq for DmInterface {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmInterface {
    pub fn get_method(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmFn>> {
        if let Some(dm_fn) = self.impl_methods.iter().find(|&dm_fn| dm_fn.fqn == name) {
            return Some(dm_fn.clone());
        } else if self
            .virtual_methods
            .iter()
            .find(|&dm_fn| dm_fn.fqn == name)
            .is_some()
        {
            return self_object.implementation.get_method(name, self_object);
        }
        None
    }

    pub fn get_property(&self, name: &str, self_object: &DmAllocObject) -> Option<&DmProperty> {
        todo!()
    }
}

#[derive(Debug, Eq)]
pub struct DmVirtualMethod {
    pub fqn: String,
    pub short_name: String,
}

impl PartialEq for DmVirtualMethod {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmImplementation {
    pub fqn: String,
    pub short_name: String,
    pub interface: Rc<DmInterface>,
    pub size_in_bytes: usize,
    properties: Vec<DmProperty>,
    fields: Vec<DmField>,
    methods: Vec<Rc<DmFn>>,
}

impl PartialEq for DmImplementation {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmImplementation {
    pub fn get_method(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmFn>> {
        for method in &self.methods {
            if method.fqn == name {
                return Some(method.clone());
            } else if let Some(implements) = &method.implements {
                if implements.fqn == name {
                    return Some(method.clone());
                }
            }
        }
        None
    }

    pub fn get_property(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmProperty>> {
        todo!()
    }

    pub fn get_field(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmField>> {
        todo!()
    }
}

#[derive(Debug, Eq)]
pub struct DmProperty {
    pub name: String,
    pub data_offset: usize,
    pub primitive_type: Rc<DmPrimitiveType>,
}

impl PartialEq for DmProperty {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

#[derive(Debug, Eq)]
pub struct DmField {
    name: String,
    primitive_type: Rc<DmPrimitiveType>,
}

impl PartialEq for DmField {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}
